import datetime
import logging
import os

from lxml import etree

from schedules_tools import models, SchedulesToolsException
from schedules_tools.schedule_handlers import (ScheduleHandlerBase,
                                               TJXCommonMixin,
                                               )

log = logging.getLogger(__name__)


class TJXImportException(SchedulesToolsException):
    pass


class ScheduleHandler_tjx(TJXCommonMixin, ScheduleHandlerBase):
    provide_changelog = True

    handle_deps_satisfied = True

    @classmethod
    def is_valid_source(cls, handle=None):
        if not handle:
            handle = cls.handle
        file_ext = os.path.splitext(handle)[1]

        if file_ext == '.tjx':
            try:
                tree = etree.parse(handle)
            except etree.XMLSyntaxError:
                return False

            if tree.xpath('//Project[@Id and @WeekStart]'):
                return True
        return False

    def import_schedule(self):
        if not self.__class__.is_valid_source(self.handle):
            raise TJXImportException('Invalid TJX source', source=self.handle)

        self.schedule = models.Schedule()

        tree = self._get_parsed_tree()

        self.schedule.name = '%s %s' % (tree.xpath('Name')[0].text.strip(),
                                        tree.xpath('Version')[0].text)

        slug = str(tree.xpath('@Id')[0])
        if slug:
            self.schedule.slug = slug

        # look for same id as project, there might be more included root tasks
        eRoot_task = None
        eRoot_tasks = tree.xpath('Task[@Id = /Project/@Id]')
        if not len(eRoot_tasks):  # try whatever single root task
            eRoot_tasks = tree.xpath('Task')
            root_tasks_count = len(eRoot_tasks)
            if root_tasks_count == 1:
                eRoot_task = eRoot_tasks[0]
            elif root_tasks_count == 0:
                log.warning('Empty schedule %s ' % (self.handle,))
        else:
            eRoot_task = eRoot_tasks[0]

        if eRoot_task is not None:
            root_task_name = eRoot_task.xpath('Name')
            if root_task_name:
                self.schedule.name = root_task_name[0].text
        else:
            log.info('Can\'t find single root task in {} (found {} root '
                     'tasks)'.format(self.handle, len(eRoot_tasks)))

        self.schedule.name = self.schedule.name.strip()

        # import changelog/mtime from content of TJX
        self.schedule.changelog = self.get_handle_changelog()
        self.schedule.mtime = self.get_handle_mtime()

        min_date = datetime.datetime.max
        max_date = datetime.datetime.min

        for eTask in eRoot_tasks:
            task = models.Task(self.schedule, level=1)
            t_min_date, t_max_date = self.task_load_tjx_node(task, eTask)
            min_date = min(min_date, t_min_date)
            max_date = max(max_date, t_max_date)
            self.schedule.tasks.append(task)

        if self.schedule.tasks:
            self.schedule.dStart = min_date
            self.schedule.dFinish = max_date
        else:
            # try to load dates from project level
            start = self._load_tjx_date(tree, 'start')
            if start:
                self.schedule.dStart = start
            end = self._load_tjx_date(tree, 'end')
            if end:
                self.schedule.dFinish = end

        if self.options.get('always_regenerate_slugs', False):
            # regenerate slugs based on name
            self.schedule.generate_slugs()

        return self.schedule

    @staticmethod
    def _load_tjx_date(eTask, datetype, what=''):
        """Returns datetime with datetype = plan|actual what = start|end"""
        tag = datetype.lower() + what.capitalize()
        eTag = eTask.xpath(tag)
        if eTag:
            return datetime.datetime.fromtimestamp(float(eTag[0].text))

    def task_load_tjx_node(self, task, eTask):
        task.index = int(eTask.xpath('Index')[0].text)
        task.slug = eTask.get('Id')
        task.name = eTask.xpath('Name')[0].text.strip()

        notes = eTask.xpath('Note')
        if notes:
            task.note = notes[0].text.strip()

        task.priority = int(eTask.xpath('Priority')[0].text)
        task.p_complete = float(eTask.xpath('complete')[0].text)
        task.dStart = self._load_tjx_date(eTask, 'actual', 'start') or self._load_tjx_date(eTask, 'plan', 'start')
        task.dFinish = self._load_tjx_date(eTask, 'actual', 'end') or self._load_tjx_date(eTask, 'plan', 'end')

        # sanity check - if only tart defined and beyond plan finish
        task.dFinish = max(task.dFinish, task.dStart)
        task.milestone = eTask.xpath('Type')[0].text == 'Milestone'

        for eFlag in eTask.xpath('./Flag'):
            task.flags.append(eFlag.text)

        task._schedule.used_flags |= set(task.flags)

        ptask_el = eTask.xpath('./custom[@id="PTask"]')
        if ptask_el:
            task.link = ptask_el[0].get('url')

        min_date = task.dStart
        max_date = task.dFinish

        for eSubTask in eTask.xpath('./SubTasks/Task'):
            task_item = models.Task(task._schedule, task.level + 1)
            t_min_date, t_max_date = self.task_load_tjx_node(task_item,
                                                             eSubTask)
            min_date = min(min_date, t_min_date)
            max_date = max(max_date, t_max_date)

            task.tasks.append(task_item)

        return min_date, max_date



import os
import sys
import re
from .schedule_handlers import ScheduleHandlerBase
import logging
import importlib

VALID_MODULE_NAME = re.compile(r'^(\w+)\.py$', re.IGNORECASE)
PARENT_DIRNAME = os.path.basename(os.path.dirname(os.path.realpath(__file__)))
BASE_DIR = os.path.dirname(os.path.realpath(
    os.path.join(__file__, os.pardir)))

log = logging.getLogger(__name__)
re_schedule_handler = re.compile(r'^ScheduleHandler_(\S+)$')
re_storage_handler = re.compile(r'^StorageHandler_(\S+)$')

# FIXME(mpavlase): Figure out nicer way to deal with paths
sys.path.append(BASE_DIR)

def get_local_path(path):
    return os.path.join(BASE_DIR, PARENT_DIRNAME, path)


class AutodiscoverHandlers(object):
    _discovered_handlers = None
    re_class_teplate = None

    def __init__(self, re_class_template):
        self.re_class_teplate = re_class_template
        self._discovered_handlers = dict()

    def get_handlers(self):
        return self._discovered_handlers

    @staticmethod
    def _load_module(name):
        importlib.import_module(name)
        return sys.modules[name]

    def _find_classes(self, module):
        ret = dict()

        for name in dir(module):
            obj = getattr(module, name)
            is_inst = isinstance(obj, type)

            if not is_inst:
                continue

            if obj == ScheduleHandlerBase:
                continue

            # TODO(mpavlase): figure out more reliable way to test subclasses
            #is_subclass = issubclass(obj, handlers.ScheduleHandlerBase)
            #if not is_subclass:
            #    continue

            key = re.findall(self.re_class_teplate, obj.__name__)
            if not key:
                continue

            key = key[0]
            val = dict()
            val['class'] = obj
            try:
                val['provide_export'] = obj.provide_export
            except AttributeError:
                # Not all handlers has this attr
                # TODO(mpavlase): consider this behavior implement in subclass
                pass

            ret[key] = val
            log.debug('Discovered new handler: {} from {}'.format(key, module))
        return ret

    def _discover_path(self, filename, parent_module):
        # valid Python identifiers only
        if not VALID_MODULE_NAME.match(filename):
            return

        name = VALID_MODULE_NAME.sub('\\1', filename)
        name = '.'.join([parent_module, name])
        loaded_module = self._load_module(name)
        classes = self._find_classes(loaded_module)

        for k in classes.keys():
            if k in self._discovered_handlers.keys():
                cls_existing = self._discovered_handlers[k]
                cls_new = classes[k]
                msg = ('Found handler with name \'{}\', that already exists '
                       'and would be overridden (existing: {}, '
                       'new: {})').format(k, cls_existing, cls_new)
                log.info(msg)

        self._discovered_handlers.update(classes)

    def discover(self, pypath):
        try:
            loaded_module = self._load_module(pypath)
        except ImportError as e:
            log.warn('Skipping path "{}", couldn\'t load'
                        'it: {} )'.format(pypath, e))
            return self._discovered_handlers

        module_path = os.path.dirname(loaded_module.__file__)
        files = os.listdir(module_path)

        for filename in files:
            self._discover_path(filename, pypath)

        return self._discovered_handlers


class LazyDictDiscovery(dict):
    autodiscovery = None
    _last_search_paths = None

    def __init__(self, *args, **kwargs):
        self.autodiscovery = AutodiscoverHandlers(kwargs.pop('cls_template'))
        self.update(dict(*args, **kwargs))  # use the free update to set keys
        super(LazyDictDiscovery, self).__init__(*args, **kwargs)

    def __getitem__(self, item):
        self.run_discovery()
        return super(LazyDictDiscovery, self).__getitem__(item)

    def keys(self):
        self.run_discovery()
        return super(LazyDictDiscovery, self).keys()

    def values(self):
        self.run_discovery()
        return super(LazyDictDiscovery, self).values()

    def _post_discovery_hook(self):
        """
        Override this method, if you need to do something after discovery has
        been processed.
        """
        pass

    def force_run_discovery(self):
        self._last_search_paths = None
        self.run_discovery()

    def run_discovery(self):
        # run only when search_paths has been changed
        if self._last_search_paths == search_paths:
            return

        ret = dict()

        for pypath in search_paths:
            log.debug('Searching for handlers in: {}'.format(pypath))
            ret = self.autodiscovery.discover(pypath)

        # override all existing keys/values
        self.clear()
        for key, value in ret.items():
            self[key] = value

        self._post_discovery_hook()
        self._last_search_paths = search_paths


class ScheduleHandlerDiscovery(LazyDictDiscovery):
    _provided_exports = []

    @property
    def provided_exports(self):
        if not self._provided_exports:
            self.run_discovery()
        return self._provided_exports

    def _post_discovery_hook(self):
        for handler_name, handler in self.items():
            if handler['provide_export']:
                self._provided_exports.append(handler_name)

        self._provided_exports = sorted(self._provided_exports)


class StorageHandlerDiscovery(LazyDictDiscovery):
    pass

schedule_handlers = ScheduleHandlerDiscovery(cls_template=re_schedule_handler)
storage_handlers = StorageHandlerDiscovery(cls_template=re_storage_handler)

search_paths = ['schedules_tools.schedule_handlers',
                'schedules_tools.storage_handlers']



from schedules_tools.schedule_handlers import ScheduleHandlerBase, TJXCommonMixin
from schedules_tools import models
import datetime
import logging
import os
from lxml import etree
from pytz import timezone

log = logging.getLogger(__name__)


class ScheduleHandler_tjx2(TJXCommonMixin, ScheduleHandlerBase):
    provide_export = False

    handle_deps_satisfied = True

    task_index = 1

    tz = timezone('America/New_York')

    @classmethod
    def is_valid_source(cls, handle=None):
        if not handle:
            handle = cls.handle
        file_ext = os.path.splitext(handle)[1]

        if file_ext == '.tjx':
            try:
                tree = etree.parse(handle)
            except etree.XMLSyntaxError:
                return False

            if tree.xpath('/taskjuggler/project[@name]'):
                return True
        return False

    def _load_tjx_date(self, eTask, datetype, what=''):
        """Returns datetime with datetype = plan|actual what = start|end as
        offset-naive datetime (without specified timezone)"""
        xpath = './taskScenario[@scenarioId=\'{}\']/{}'.format(datetype, what)
        tag = eTask.xpath(xpath)
        if tag:
            date = datetime.datetime.fromtimestamp(float(tag[0].text),
                                                   tz=self.tz)
            # Parsed timestamps are timesone-aware, but we don't want to store
            # this awareness.
            return date.replace(tzinfo=None)
        else:
            return None

    def _parse_task_element(self, task, eTask):
        task.index = self.task_index
        self.task_index += 1
        task.slug = eTask.get('id')
        task.name = eTask.get('name').strip()

        notes = eTask.xpath('note')
        if notes:
            task.note = notes[0].text.strip()

        task.priority = int(eTask.get('priority'))
        task.milestone = eTask.get('milestone') == '1'

        p_complete_attr = eTask.xpath('./taskScenario[@scenarioId=\'actual\']')
        p_complete = float(p_complete_attr[0].get('complete'))
        if p_complete < 0:
            p_complete = 0.0
        task.p_complete = p_complete

        task.dStart = self._load_tjx_date(eTask, 'actual', 'start') or self._load_tjx_date(eTask, 'plan', 'start')
        task.dFinish = self._load_tjx_date(eTask, 'actual', 'end') or self._load_tjx_date(eTask, 'plan', 'end')

        if task.milestone:
            task.dFinish = task.dStart

        for eFlag in eTask.xpath('./flag'):
            task.flags.append(eFlag.text)

        link_el = eTask.xpath('customAttribute[@id="PTask"]/referenceAttribute')
        if link_el:
            task.link = link_el[0].get('url')

        # add flags from task to global used tags
        task._schedule.used_flags |= set(task.flags)

        min_date = task.dStart
        max_date = task.dFinish

        for eSubTask in eTask.xpath('./task'):
            item_task = models.Task(self.schedule, task.level + 1)
            t_min_date, t_max_date = self._parse_task_element(item_task,
                                                              eSubTask)
            min_date = min(min_date, t_min_date)
            max_date = max(max_date, t_max_date)

            task.tasks.append(item_task)
        return min_date, max_date

    def import_schedule(self):
        self.schedule = models.Schedule()

        tree = self._get_parsed_tree()
        el_proj = tree.xpath('/taskjuggler/project')[0]

        self.schedule.name = '%s %s' % (el_proj.get('name'),
                                        el_proj.get('version'))
        self.schedule.slug = el_proj.get('id')

        # import changelog/mtime from content of TJX
        self.schedule.changelog = self.get_handle_changelog()
        self.schedule.mtime = self.get_handle_mtime()

        min_date = datetime.datetime.max
        max_date = datetime.datetime.min

        for task in tree.xpath('/taskjuggler/taskList/task'):
            item_task = models.Task(self.schedule)
            self._parse_task_element(item_task, task)
            min_date = min(min_date, item_task.dStart)
            max_date = max(max_date, item_task.dFinish)
            self.schedule.tasks.append(item_task)

        if self.schedule.tasks:
            self.schedule.dStart = min_date
            self.schedule.dFinish = max_date
            self.schedule.name = self.schedule.tasks[0].name
        else:
            # try to load dates from project level
            tag = el_proj.xpath('start')[0].text
            start = datetime.datetime.fromtimestamp(float(tag))
            if start:
                self.schedule.dStart = start

            tag = el_proj.xpath('end')[0].text
            end = datetime.datetime.fromtimestamp(float(tag))
            if end:
                self.schedule.dFinish = end
        self.schedule.name = self.schedule.name.strip()

        if self.options.get('always_regenerate_slugs', False):
            # regenerate slugs based on name
            self.schedule.generate_slugs()

        return self.schedule
